/* 
 * version 3
 *
 * This code is copyright Thouis Jones, 
 * and distributed under the terms of the GPLv2
 */

#include <gts.h>
#include <math.h>
#include <time.h>
#include <stdlib.h>

static GHashTable *mollified_hash, *new_position_hash;
static gint num_verts;
static GNode *tree;
static gdouble sigma_f, sigma_g, sigma_m, cutoff;

gdouble gaussian2(double r, double sigma)
{
  /* don't worry about the 1/(sigma*sqrt(2*PI)) normalization for now,
   * as there is a separate normalization in every function that uses
   * this.  But be aware that this may change. */

  return (exp(- r / (2 * sigma * sigma)));
}

gint
box_by_centroid(gpointer item, gpointer data)
{
  GtsTriangle *t = item;
  GSList *l = * (GSList **) data;
  GtsVertex *v1, *v2, *v3;
  gdouble centroid[3];
  
  gts_triangle_vertices(t, &v1, &v2, &v3);
  centroid[0] = (GTS_POINT(v1)->x + 
                 GTS_POINT(v2)->x + 
                 GTS_POINT(v3)->x) / 3.0;
  centroid[1] = (GTS_POINT(v1)->y + 
                 GTS_POINT(v2)->y + 
                 GTS_POINT(v3)->y) / 3.0;
  centroid[2] = (GTS_POINT(v1)->z + 
                 GTS_POINT(v2)->z + 
                 GTS_POINT(v3)->z) / 3.0;
  
  l = g_slist_prepend(l, gts_bbox_new(gts_bbox_class(),
                                      t,
                                      centroid[0], centroid[1], centroid[2],
                                      centroid[0], centroid[1], centroid[2]));
  * (GSList **) data = l;

  return (0);
}

void
project_to_tri(GtsVertex *p,
               gdouble centroid[3],
               GtsVertex *v1, GtsVertex *v2, GtsVertex *v3,
               gdouble *x, gdouble *y, gdouble *z)
{
  GtsVector vcp, v12, v13, normal;
  gdouble dotvcpn;
  
  vcp[0] = GTS_POINT(p)->x - centroid[0];
  vcp[1] = GTS_POINT(p)->y - centroid[1];
  vcp[2] = GTS_POINT(p)->z - centroid[2];
  gts_vector_init(v12, GTS_POINT(v1), GTS_POINT(v2));
  gts_vector_init(v13, GTS_POINT(v1), GTS_POINT(v3));
  
  gts_vector_cross(normal, v12, v13);
  gts_vector_normalize(normal);

  dotvcpn = gts_vector_scalar(vcp, normal);

  *x = GTS_POINT(p)->x - dotvcpn * normal[0];
  *y = GTS_POINT(p)->y - dotvcpn * normal[1];
  *z = GTS_POINT(p)->z - dotvcpn * normal[2];
}

#define DIST(a,b,c) ((a)*(a)+(b)*(b)+(c)*(c))

void
filter(GtsVertex *cur_vert, GtsVertex *weighted_pos, gdouble *k,
       GNode *tree,
       gdouble spatial_sigma, gdouble influence_sigma, 
       GHashTable *vertex_map)
{
  gdouble dmin, dmax;

  gts_bbox_point_distance2(GTS_BBOX(tree->data), GTS_POINT(cur_vert),
                           &dmin, &dmax);
  if (dmin > (cutoff * cutoff)) return;

  if (G_NODE_IS_LEAF(tree)) {
    GtsTriangle *t = GTS_BBOX(tree->data)->bounded;
    GtsVertex *v1, *v2, *v3;
    gdouble centroid[3];
    gdouble px, py, pz;
    gdouble w = 0.0;
    gdouble area;

    area = gts_triangle_area(t);

    gts_triangle_vertices(t, &v1, &v2, &v3);
    
    /* calculate centroid from original vertex positions */
    centroid[0] = (GTS_POINT(v1)->x + 
                   GTS_POINT(v2)->x + 
                   GTS_POINT(v3)->x) / 3.0;
    centroid[1] = (GTS_POINT(v1)->y + 
                   GTS_POINT(v2)->y + 
                   GTS_POINT(v3)->y) / 3.0;
    centroid[2] = (GTS_POINT(v1)->z + 
                   GTS_POINT(v2)->z + 
                   GTS_POINT(v3)->z) / 3.0;

    if (vertex_map) {
      v1 = g_hash_table_lookup(vertex_map, v1);
      v2 = g_hash_table_lookup(vertex_map, v2);
      v3 = g_hash_table_lookup(vertex_map, v3);
    }

    /* project to triangle based on original centroid position, but mollified normal */
    project_to_tri(cur_vert,
                   centroid,
                   v1, v2, v3,
                   &px, &py, &pz);

    if (influence_sigma > 0.0) {
      gdouble pdist = DIST(px - GTS_POINT(cur_vert)->x,
                           py - GTS_POINT(cur_vert)->y,
                           pz - GTS_POINT(cur_vert)->z);
      w = area * gaussian2(dmin, spatial_sigma) * gaussian2(pdist, influence_sigma);
    } else {
      w = area * gaussian2(dmin, spatial_sigma);
    }

    GTS_POINT(weighted_pos)->x += w * px;
    GTS_POINT(weighted_pos)->y += w * py;
    GTS_POINT(weighted_pos)->z += w * pz;
    *k += w;

  } else {
    filter(cur_vert, weighted_pos, k,
           g_node_first_child(tree),
           spatial_sigma, influence_sigma,
           vertex_map);
    filter(cur_vert, weighted_pos, k,
           g_node_nth_child(tree, 1),
           spatial_sigma, influence_sigma,
           vertex_map);
  }
}
       

gint
mollify_vertex(gpointer item, gpointer data)
{
  GtsVertex *v = item, *new_pos;
  gdouble k;

  static gint vnum = 0;
  gint i;

  /* status bar */
  vnum++;
  if (!(vnum % 1000)) {
    fprintf(stderr, "Mollifying:[");
    for(i = 0; i < (65 * vnum) / num_verts; i++) 
      fprintf(stderr, "#");
    for(; i < 65; i++) 
      fprintf(stderr, " ");
    fprintf(stderr, "]\r");
  }
  if (vnum == num_verts) fprintf(stderr, "\n");

  new_pos = gts_vertex_new(gts_vertex_class(),
                           0.0, 0.0, 0.0);
  k = 0.0;
  filter(v, new_pos, &k, tree, sigma_m, 0.0, NULL);

  if (k > 0.0) {
    GTS_POINT(new_pos)->x /= k;
    GTS_POINT(new_pos)->y /= k;
    GTS_POINT(new_pos)->z /= k;
  } else {
    GTS_POINT(new_pos)->x = GTS_POINT(v)->x;
    GTS_POINT(new_pos)->y = GTS_POINT(v)->y;
    GTS_POINT(new_pos)->z = GTS_POINT(v)->z;
  }

  g_hash_table_insert(mollified_hash, v, new_pos);

  return (0);
}

gint
filter_vertex(gpointer item, gpointer data)
{
  GtsVertex *v = item, *new_pos;
  gdouble k;

  static gint vnum = 0;
  gint i;

  /* status bar */
  vnum++;
  if (!(vnum % 1000)) {
    fprintf(stderr, "Moving:[");
    for(i = 0; i < (65 * vnum) / num_verts; i++) 
      fprintf(stderr, "#");
    for(; i < 65; i++) 
      fprintf(stderr, " ");
    fprintf(stderr, "]\r");
  }
  if (vnum == num_verts) fprintf(stderr, "\n");

  new_pos = gts_vertex_new(gts_vertex_class(),
                           0.0, 0.0, 0.0);
  k = 0.0;
  filter(v, new_pos, &k, tree, sigma_f, sigma_g, mollified_hash);

  if (k > 0.0) {
    GTS_POINT(new_pos)->x = GTS_POINT(new_pos)->x / k;
    GTS_POINT(new_pos)->y = GTS_POINT(new_pos)->y / k;
    GTS_POINT(new_pos)->z = GTS_POINT(new_pos)->z / k;
  } else {
    GTS_POINT(new_pos)->x = GTS_POINT(v)->x;
    GTS_POINT(new_pos)->y = GTS_POINT(v)->y;
    GTS_POINT(new_pos)->z = GTS_POINT(v)->z;
  }
  
  g_hash_table_insert(new_position_hash, v, new_pos);

  return (0);
}

gint
move_vertex(gpointer item, gpointer data)
{
  GtsVertex *v = item, *new_pos;

  new_pos = g_hash_table_lookup(new_position_hash, v);

  GTS_POINT(v)->x = GTS_POINT(new_pos)->x;
  GTS_POINT(v)->y = GTS_POINT(new_pos)->y;
  GTS_POINT(v)->z = GTS_POINT(new_pos)->z;

  return (0);
}

int main (int argc, char * argv[])
{
  GtsFile *fp;
  GtsSurface *s;
  GtsSurfaceQualityStats qstats;
  time_t start;
  GSList *trilist = NULL;

  if (argc != 3) {
    fprintf(stderr, "Usage %s sigma_f sigma_g < in.gts > out.gts\n",
            argv[0]);
    exit (-1);
  }

  /* read surface in */
  s = gts_surface_new (gts_surface_class (),
		       gts_face_class (),
		       gts_edge_class (),
		       gts_vertex_class ());
  fp = gts_file_new (stdin);
  if (gts_surface_read (s, fp)) {
    fprintf(stderr, "input not a valid GTS file\n");
    return 1; /* failure */
  }

  gts_surface_quality_stats(s, &qstats);
  gts_surface_print_stats(s, stderr);

  // read command line arguments
  sigma_f = atof(argv[1]);
  sigma_g = atof(argv[2]);
  printf("# sigma_f %f   sigma_g %f\n", sigma_f, sigma_g);

  // scale by mean edge length
  sigma_f *= qstats.edge_length.mean;
  sigma_g *= qstats.edge_length.mean;

  cutoff = 2.0 * sigma_f;
  sigma_m = sigma_f / 2.0;

  gts_surface_foreach_face(s, box_by_centroid, &trilist);
  tree = gts_bb_tree_new(trilist);

  mollified_hash = g_hash_table_new(NULL, NULL);
  new_position_hash = g_hash_table_new(NULL, NULL);
  num_verts = gts_surface_vertex_number(s);

  start = time(NULL);
  gts_surface_foreach_vertex(s, mollify_vertex, NULL);
  gts_surface_foreach_vertex(s, filter_vertex, NULL);
  gts_surface_foreach_vertex(s, move_vertex, NULL);
  printf("# time taken: %d secs\n", (guint) (time(NULL) - start));

  gts_surface_write(s, stdout);

  return 0; /* success */
}
